import pytest

import odoo_connect


def test_connect_and_check_user(connect_params):
    env = odoo_connect.connect(**connect_params)
    assert env.protocol == 'jsonrpc'
    user = env.user
    print(user)
    assert isinstance(user, dict)
    assert user['login'] == connect_params['username']


def test_database(odoo_cli):
    assert odoo_cli.database == 'odoo', "Odoo should be the default database"


@pytest.mark.tryfirst
def test_version(connect_params):
    env = odoo_connect.connect(connect_params['url'], 'default')
    version = env.version()
    print(version)
    assert isinstance(version, dict)


def test_list_databases(odoo_cli, odoo_json_rpc_handler):
    @odoo_json_rpc_handler.patch_generic
    def list_db(service, method, args):
        if service == 'db' and method == 'list':
            return ['odoo']

    databases = odoo_cli.list_databases()
    assert databases == ['odoo']


def test_model_read(odoo_cli):
    users = odoo_cli.get_model('res.users')
    assert users and isinstance(odoo_cli['res.users'], type(users))
    data = users.read(1)
    assert isinstance(data, list) and len(data)



import json
import typing

import pytest
from werkzeug.wrappers import Request, Response


class OdooMockedError(Exception):
    """Exception not reraised during a mocked rpc call"""

    pass


class OdooRPCHandler:
    """Used as a handler for Requests"""

    def __init__(self):
        self.call_generic = []
        self.call_execute_kw = []

    def execute_kw(
        self, database: str, uid: int, password: str, model: str, function: str, a: list, kw: dict
    ):
        """Implement execute_kw behaviour, just checks that the user is authenticated (any uid)"""
        assert database and uid > 0 and password, "Must be authenticated"
        for f in self.call_execute_kw:
            r = f(model, function, a, kw)
            if r is not None:
                return r
        pytest.fail('execute_kw not implemented for %s.%s' % (model, function))

    def generic(self, service: str, method: str, args: list):
        """Implement the call method from odoo"""
        if service == 'object' and method == 'execute_kw':
            return self.execute_kw(*args)
        for f in self.call_generic:
            r = f(service, method, args)
            if r is not None:
                return r
        pytest.fail('%s.%s not implemented' % (service, method))

    def patch_generic(self, f):
        """Append a method to `generic` call.

        If the method returns None, another one is tried.

        :param f: f(service, method, args) used for generic calls
        """
        self.call_generic.append(f)
        return f

    def patch_execute_kw(self, model, function):
        """Append a method to `execute_kw` call.

        If the method returns None, another one is tried.

        :param model: Model for the object's execute_kw
        :param function: Function name
        :return: Decorator for a function call where arguments will be expanded
        """

        def _patch_execute_kw(f):
            def checked_f(model_, function_, a, kw):
                if model == model_ and function == function_:
                    return f(*a, **kw)

            self.call_execute_kw.append(checked_f)
            return checked_f

        return _patch_execute_kw

    def __call__(self, request: Request):
        """Handle a jsonrpc request for the method call"""
        assert request.content_type == "application/json"
        data = typing.cast(dict, request.json)
        assert data["jsonrpc"] == "2.0" and data["id"]
        assert data["method"] == "call", "Only call method is implemented"
        status = 200
        output = {
            "jsonrpc": "2.0",
            "id": data["id"],
        }
        try:
            params = data["params"]
            output["result"] = self.generic(**params)
        except Exception as e:
            status = 500
            output["error"] = str(e)
            if not isinstance(e, OdooMockedError):
                raise  # don't try to transform errors during testing
        return Response(json.dumps(output), status=status, content_type="application/json")


def default_rpc_handler():
    """Instantiates a default handler. Serves as an example code too."""
    h = OdooRPCHandler()

    @h.patch_generic
    def version(service, method, args):
        if service == 'common' and method == 'version':
            return {"server_version": "15.0", "server_serie": "mocked"}

    @h.patch_generic
    def authenticate(service, method, args):
        if method == 'login':
            args = args + [{}]
            method = 'authenticate'
        if service == 'common' and method == 'authenticate':
            database, username, password, env = args
            if username == 'admin' and password == 'admin':
                return 1
            if username and username == password:
                return 2
            raise OdooMockedError('Cannot authenticate on %s with %s' % (database, username))

    @h.patch_execute_kw('res.users', 'read')
    def read_user(id, fields=[]):
        if not (0 < id < 10):
            return None  # not implemented for other id's
        result = {
            'id': id,
            'login': 'admin' if id == 1 else 'other',
            'name': 'some user',
            'parnter_id': [id, 'Contact'],
        }
        if fields:
            fields = ['id'] + fields
            result = {k: v for k, v in result.items() if k in fields}
        return [result]

    return h



import odoo_connect


def test_user_with_protocol(connect_params):
    env = odoo_connect.connect(**connect_params)
    user = env.user
    print(user)
    assert isinstance(user, dict), "Failed to get a response"
    assert user['login'] == connect_params['username']


def test_version(connect_params):
    env = odoo_connect.connect(connect_params['url'])
    version = env.version()
    print(version)
    assert isinstance(version, dict)


def test_list_models(odoo_session):
    models = odoo_session.list_models()
    assert isinstance(models, list) and 'res.users' in models


def test_model_search_user(odoo_session):
    data = odoo_session['res.users'].search_read(
        [('active', 'in', [True, False])], ['login'], limit=2
    )
    print(data)
    assert len(data) == 2, "There should always be at least 2 users"
    assert data[0]['login'], "Every user must have a login"


def test_ref(odoo_session):
    data = odoo_session.ref('base.group_user', fields=['name'])
    print(data)
    assert isinstance(data, dict)



import base64
from collections import defaultdict
from datetime import date, datetime
from typing import Any, Callable, Dict, Optional, Union, cast

from .odoo_rpc import OdooModel

__doc__ = """Formatting of fields for Odoo."""

NOT_FORMATTED_FIELDS = {
    'display_name',
    'write_date',
    'create_date',
    'write_uid',
    'create_uid',
    '__last_update',
}

"""Default formatters for models"""
DEFAULT_FORMATTERS: Dict[OdooModel, "Formatter"] = {}


def decode_default(value) -> Any:
    """Decode a value from Odoo (identity function)"""
    return value


def decode_datetime(value) -> Optional[datetime]:
    if not value:
        return None
    return datetime.fromisoformat(value)


def decode_date(value) -> Optional[date]:
    if not value:
        return None
    return date.fromisoformat(value[:10])


def decode_binary(value) -> bytes:
    """Decode bytes from a base64"""
    if not value:
        return b''
    return base64.b64decode(value)


def format_default(v: Any):
    """Format a value for Odoo"""
    if isinstance(v, str):
        v = v.strip()
    return v or False


def format_datetime(v: Union[datetime, date, str]):
    """Format a datetime to insert into Odoo"""
    if isinstance(v, datetime):
        # format ISO with a space an ignore millisenconds
        return v.isoformat(sep=" ", timespec='seconds')
    if isinstance(v, str):
        return v.strip()[:19] or False
    if isinstance(v, date):
        return v.isoformat() + " 00:00:00"
    return False


def format_date(v: Union[datetime, date, str]):
    """Format a date to insert into Odoo"""
    if isinstance(v, datetime):
        v = v.date()
    if isinstance(v, date):
        return v.isoformat()
    if isinstance(v, str):
        return v.strip()[:10] or False
    return False


def format_binary(v: Union[bytes, str]) -> str:
    """Format bytes to insert into Odoo (as base64)"""
    if isinstance(v, str):
        v = bytes(v, 'utf-8')
    encoded = base64.b64encode(v)
    return str(encoded, 'ascii')


class Formatter:
    """Format fields into a format expected by Odoo."""

    """Transformations to apply to source fields.
    Use an empty string to mask some of them."""
    field_map: Dict[str, str]
    format_function: Dict[str, Callable[[Any], Any]]
    decode_function: Dict[str, Callable[[Any], Any]]

    def __init__(self, model: OdooModel = None, *, lower_case_fields: bool = False):
        """New formatter

        :param model: The model for which the formatter is generated (initializes formatters)
        :param lower_case_fields: Whether to transform source fields into lower-case
        """
        self.format_function = defaultdict(lambda: format_default)
        self.decode_function = defaultdict(lambda: decode_default)
        self.model = model
        self.lower_case_fields = lower_case_fields
        self.field_map = {}
        if not model:
            return
        # set the formats from the model
        functions = {
            'datetime': (format_datetime, decode_datetime),
            'date': (format_date, decode_date),
            'binary': (format_binary, decode_binary),
            'image': (format_binary, decode_binary),
        }
        for field, info in model.fields().items():
            formatter, decoder = functions.get(
                cast(str, info.get('type')), (format_default, decode_default)
            )
            if formatter is not format_default:
                self.format_function.setdefault(field, formatter)
            if decoder is not decode_default:
                self.decode_function.setdefault(field, decoder)

    def map_field_name(self, name: str) -> str:
        """Transform a source field name into model's name"""
        name = self.field_map.get(name, name)
        if self.lower_case_fields:
            name = name.lower()
        return name

    def format_dict(self, d: Dict[str, Any]) -> Dict[str, Any]:
        """Apply formatting to each field"""
        d_renamed = [(self.map_field_name(f), v) for f, v in d.items()]
        return {
            f: self.format_function[f](v)
            for f, v in d_renamed
            if f and f not in NOT_FORMATTED_FIELDS
        }

    def decode_dict(self, d: Dict[str, Any]) -> Dict[str, Any]:
        """Apply decoding to each field"""
        return {f: self.decode_function[f](v) for f, v in d.items()}


def get_default_formatter(model: OdooModel) -> Formatter:
    """Get the default dict formatter function"""
    formatter = DEFAULT_FORMATTERS.get(model)
    if not formatter:
        DEFAULT_FORMATTERS[model] = formatter = Formatter(model)
    return formatter



import pathlib

import pytest

from . import mock_odoo_server


@pytest.fixture(scope='function')
def connect_params(httpserver, odoo_json_rpc_handler):
    # use odoo_json_rpc_handler so it is set up
    return {
        'url': httpserver.url_for('/'),
        'username': 'admin',
        'database': 'odoo',
    }


@pytest.fixture(scope='function')
def odoo_json_rpc_handler(httpserver):
    """Setup the http server for Odoo JSON RPC"""
    handler = mock_odoo_server.default_rpc_handler()
    httpserver.expect_request(
        "/jsonrpc", headers={'content-type': 'application/json'}
    ).respond_with_handler(handler)
    return handler


@pytest.fixture(scope='function')
def odoo_cli(connect_params):
    import odoo_connect

    return odoo_connect.connect(**connect_params)


# CONFIGURE PYTEST


def pytest_configure(config):
    config.addinivalue_line("markers", "runbot: integration tests on runbot")


def pytest_collection_modifyitems(config, items):
    # If a test is in a subdirectory, add marker which is the directory name
    # https://stackoverflow.com/questions/57031403/pytest-marks-mark-entire-directory-package
    # To mark a file, you can use pytestmark = pytest.mark.my_mark
    rootdir = pathlib.Path(config.rootdir)
    for item in items:
        rel_path = pathlib.Path(item.fspath).relative_to(rootdir)
        mark_name = next((part for part in rel_path.parts if not part.startswith('test')), '')
        if mark_name:
            mark = getattr(pytest.mark, mark_name)
            item.add_marker(mark)



import urllib.parse
from typing import Optional

from .odoo_rpc import OdooClient, OdooModel, OdooServerError  # noqa

__doc__ = """Simple Odoo RPC library."""


class OdooConnectionError(OdooServerError):
    """Connection error"""

    pass


def connect(
    url: str,
    database: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    infer_parameters: bool = True,
    check_connection: bool = True,
) -> OdooClient:
    """Connect to an odoo database.

    When infer_paramters is set, the url is parsed to get additional information.
    - When missing, the scheme is https (or http on localhost)
    - When missing we try the following heuristics: the database is read from the path,
      in a multipart host name, the first part is used, otherwise a default database
    - The username and password are read if present in the url
    - When missing, the password is copied from the user

    Some examples for infered parameters:
    - https://user:pwd@hostname/database
    - mytest.odoo.com -> https://mytest.odoo.com/mytest
    - localhost -> http://localhost/odoo
    - https://admin@myserver:8069 would connect with password "admin" to the default database

    :param url: The URL to the server, it may encode other information when infer_parameters is set
    :param database: The database name
    :param username: The username (when set, we try to authenticate the user during the connection)
    :param password: The password
    :param infer_paramters: Whether to infer parameters (default: True)
    :param check_connection: Try to connect (default: True)
    :return: Connection object to the Odoo instance
    """
    urlx = urllib.parse.urlparse(url)
    if infer_parameters:
        if not urlx.scheme and not urlx.netloc and urlx.path:
            # we just have a server name in the path (reparse with slashes)
            urlx = urllib.parse.urlparse('//' + urlx.path.lstrip('/'))
        if not urlx.hostname:
            raise ValueError(f"No hostname in url {url}")
        if not database and len(urlx.path) > 1:
            # extract the database from the path if it's there
            path = urlx.path.lstrip('/')
            if '/' not in path:
                database = path
                urlx = urlx._replace(path='/')
        if not database:
            # try to extract the database from the hostname
            # dbname.runbot*.odoo.com or dbname.dev.odoo.com
            name_split = (urlx.hostname or '').split('.')
            if len(name_split) > 3:
                database = name_split[0]
        if not username and urlx.username:
            # read username and password from the url
            username = urlx.username
            password = urlx.password
        if not password and username:
            # copy username to password when not set
            password = username
        # make sure the url does not contain credentials anymore
        at_loc = urlx.netloc.find('@')
        if at_loc > 0:
            urlx = urlx._replace(netloc=urlx.netloc[at_loc + 1 :])
    if not urlx.scheme:
        # add a scheme
        urlx = urlx._replace(scheme="http" if urlx.hostname == "localhost" else "https")
    url = urlx.geturl()

    # Create the connection
    try:
        client = OdooClient(url=url, database=database)
        if username:
            client.authenticate(username, password or '')
        elif check_connection:
            client.version()
        return client
    except (NotImplementedError, OdooConnectionError):
        raise
    except (ConnectionError, IOError, OdooServerError) as e:
        raise OdooConnectionError(e)



import logging
import random
import re
from typing import Any, Dict, List, Optional, Union

import requests

__doc__ = """
JSON-RPC class for Odoo RPC.
"""


def urljoin(base: str, *parts) -> str:
    """Simple URL joining"""
    if not parts:
        return base
    if base.endswith("/"):
        base = base[:-1]
    return "/".join([base] + [p.strip("/") for p in parts])


def get_month(value: str) -> int:
    """Get the month number from a month name"""
    month = value.lower()[:3]
    return {
        'jan': 1,
        'feb': 2,
        'mar': 3,
        'apr': 4,
        'may': 5,
        'jun': 6,
        'jul': 7,
        'aug': 8,
        'sep': 9,
        'oct': 10,
        'nov': 11,
        'dec': 12,
    }.get(month, 0)


class OdooServerError(RuntimeError):
    """Error returned by Odoo"""

    def get_data(self) -> Optional[dict]:
        """Get the data dictionnary for the error"""
        return next((a for a in self.args if isinstance(a, dict)), None)

    def get_remote_trace(self) -> Optional[str]:
        """Get the debug trace received from the remote server"""
        data = self.get_data() or {}
        dat = data.get('data') or {}
        return dat.get('debug')


class OdooClient:
    """Odoo server connection"""

    url: str
    _models: Dict[str, "OdooModel"]
    _version: Dict[str, Any]
    _database: str
    _username: str
    _password: str
    _uid: Optional[int]

    def __init__(
        self,
        *,
        url: str,
        database: Optional[str] = None,
        **_kwargs,
    ):
        """Create new connection."""
        self.url = url
        self._database = database or ''
        self._models = {}
        self._version = {}
        self._username = ''
        self._password = ''
        self._uid = None
        self._init_session()
        if not database:
            self._database = self._init_default_database()
        assert self._database
        logging.getLogger(__name__).info(
            "Odoo connection initialized [%s], db: [%s]",
            self.url,
            self.database,
        )

    def _init_session(self):
        """Initialize the session"""
        self.__json_url = urljoin(self.url, "jsonrpc")
        self.session = requests.Session()

    def _init_default_database(self) -> str:
        """Gets the default database from the server"""
        log = logging.getLogger(__name__)
        log.debug("Lookup the default database for [%s]", self.url)
        # Get the default
        try:
            db = self._call("db", "monodb")
            if isinstance(db, str) and db:
                return db
        except OdooServerError:
            pass
        # Try to list databases
        dbs = self.list_databases()
        if len(dbs) == 1:
            return dbs[0]
        # Fail
        raise OdooServerError('Cannot determine the database for [%s]' % self.url)

    def authenticate(self, username: str, password: str):
        """Authenticate with username and password"""
        log = logging.getLogger(__name__)
        old_username = self._username
        self._uid = None
        self._username = username
        self._password = password
        if not username:
            if old_username:
                log.info('Logged out [%s]' % self.url)
            return
        user_agent_env = {}  # type: ignore
        self._uid = self._call(
            "common",
            "authenticate",
            self._database,
            self._username,
            self._password,
            user_agent_env,
        )
        if not self._uid:
            raise OdooServerError('Failed to authenticate user %s' % username)
        log.info("Login successful [%s], [%s] uid: %d", self.url, self.username, self._uid)

    def _json_rpc(self, method: str, params: Any):
        """Make a jsonrpc call"""
        data = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params,
            "id": random.randint(0, 1000000000),
        }
        resp = self.session.post(self.__json_url, json=data)
        resp.raise_for_status()
        reply = resp.json()
        if reply.get("error"):
            raise OdooServerError(reply["error"])
        return reply.get("result", None)

    def _call(self, service: str, method: str, *args):
        return self._json_rpc("call", {"service": service, "method": method, "args": args})

    def _execute_kw(self, model: str, method: str, *args, **kw):
        """Execute a method on a model"""
        if not self._uid:
            raise RuntimeError('You must authenticate first')
        return self._call(
            "object",
            "execute_kw",
            self._database,
            self._uid,
            self._password,
            model,
            method,
            args,
            kw,
        )

    def get_model(self, model_name: str, check: bool = False) -> "OdooModel":
        """Get a model instance

        :param model: Name of the model
        :param check: Check if the model exists (default: no), if doesn't exist, return None
        :return: Proxy for the model functions
        """
        model = self._models.get(model_name)
        if model is None:
            model = OdooModel(self, model_name)
            self._models[model_name] = model
        if check:
            try:
                # call any method to check if the call works
                # let's fetch the fields (which we probably will do anyways)
                model.fields()
            except:  # noqa: E722
                raise OdooServerError('Model %s not found' % model)
        return model

    def list_databases(self) -> List[str]:
        """Get the list of databases (may be disabled on the server and fail)"""
        return self._call("db", "list")

    def list_models(self) -> List[str]:
        """Get the list of known model names."""
        models = self.get_model('ir.model').search_read([], ['model'])
        return [m['model'] for m in models]

    def ref(
        self, xml_id: str, fields: List[str] = [], raise_if_not_found: bool = True
    ) -> Optional[Dict]:
        """Read the record corresponding to the given `xml_id`."""
        if '.' not in xml_id:
            raise ValueError('xml_id not valid')
        module, name = xml_id.split('.', 1)
        rec = self.get_model('ir.model.data').search_read(
            [('module', '=', module), ('name', '=', name)], ['id', 'model', 'res_id'], limit=1
        )

        if rec:
            rec = rec[0]
            model = self.get_model(rec.get('model'))
            to_return = model.read(rec.get('res_id'), fields)
            if to_return:
                return to_return[0]
        if raise_if_not_found:
            raise ValueError(
                'No record found for unique ID %s. It may have been deleted.' % (xml_id)
            )
        return {}

    def version(self) -> dict:
        """Get the version information from the server"""
        if self._version:
            return self._version
        self._version = self._call(
            "common",
            "version",
        )
        return self._version

    @property
    def major_version(self) -> int:
        return self.version()['server_version_info'][0]

    @property
    def protocol(self) -> str:
        """Get protocol used"""
        return "jsonrpc"

    def is_connected(self) -> bool:
        """Check if the authentication is done"""
        return self._uid is not None

    @property
    def username(self) -> str:
        """Get username"""
        return self._username

    @property
    def user(self) -> dict:
        """Get user information"""
        if not self.is_connected:
            return {}
        data = self.get_model('res.users').read(
            self._uid,
            [
                'login',
                'name',
                'groups_id',
                'partner_id',
                'login_date',
            ],
        )
        return data[0] if data else None

    @property
    def database(self) -> str:
        """Get database name"""
        return self._database

    @database.setter
    def database(self, database: str):
        if database is None:
            raise ValueError('Cannot set database: None')
        self.authenticate('', '')  # log out first
        self._database = database

    def __getitem__(self, model: str) -> "OdooModel":
        """Alias for get_model"""
        return self.get_model(model)

    def __repr__(self) -> str:
        user = str(self._uid or self._username)
        return f"OdooClient({self.url},{self.protocol},db:{self.database},user:{user})"


class OdooModel:
    """Odoo model (object) RPC functions"""

    def __init__(self, odoo: OdooClient, model: str):
        """Initialize the model instance.

        :param odoo: Odoo instance
        :param model: Name of the model
        """
        self.odoo = odoo
        self.model = model
        self._field_info = None

    def __getattr__(self, name: str):
        """By default, return function bound to execute(name, ...)"""

        def odoo_wrapper(*args, **kw):
            return self.execute(name, *args, **kw)

        return odoo_wrapper

    def execute(self, method: str, *args, **kw):
        """Execute an rpc method with arguments"""
        logging.getLogger(__name__).debug("Execute %s on %s", method, self.model)
        return self.odoo._execute_kw(
            self.model,
            method,
            *args,
            **kw,
        )

    def __repr__(self) -> str:
        return repr(self.odoo) + "/" + self.model

    def fields(self, extended=False) -> Dict[str, dict]:
        """Returns the fields of the model"""
        if not self._field_info or (extended and not self._field_info['id'].get('name')):
            attributes = (
                [] if extended else ['string', 'type', 'readonly', 'required', 'store', 'relation']
            )
            self._field_info = self.execute(
                'fields_get',
                allfields=[],
                attributes=attributes,
            )
        return self._field_info  # type: ignore

    def __prepare_dict_fields(self, fields: Union[List[str], Dict[str, Dict]]) -> Dict[str, Dict]:
        """Make sure fields is a dict representing the data to get"""
        if isinstance(fields, list):
            new_fields: Dict[str, Dict] = {}
            for field in fields:
                level = new_fields
                for f in field.split('.'):
                    if f not in level:
                        level[f] = {}
                    level = level[f]
            return new_fields
        if isinstance(fields, dict):
            new_fields = {}
            for k, v in fields.items():
                if isinstance(v, set):
                    v = list(v)
                if isinstance(v, list):
                    new_fields[k] = self.__prepare_dict_fields(v)
            if new_fields:
                new_fields.update({k: v for k, v in fields.items() if k not in new_fields})
                return new_fields
            return fields
        raise ValueError('Invalid fields parameter: %s' % fields)

    def __read_dict_date(self, data, fields):
        """Transform dates into ISO-like format"""
        for field in fields:
            mapper = None
            if field.endswith(':quarter'):
                regex = re.compile(r'Q(\d) (\d+)')

                def mapper(v, range):
                    m = v and regex.match(v)
                    return "%s-Q%d" % (m.group(2), int(m.group(1))) if m else v

            elif field.endswith(':month'):
                regex = re.compile(r'(\w+) (\d+)')

                def mapper(v, range):
                    m = v and regex.match(v)
                    return "%s-%02d" % (m.group(2), get_month(m.group(1))) if m else v

            elif field.endswith(':week'):
                regex = re.compile(r'W(\w+) (\d+)')

                def mapper(v, range):
                    m = v and regex.match(v)
                    return "%s-W%02d" % (m.group(2), int(m.group(1))) if m else v

            elif field.endswith(':day'):
                regex = re.compile(r'(\d+) (\w+) (\d+)')

                def mapper(v, range):
                    m = v and regex.match(v)
                    return (
                        "%s-%02d-%02d" % (m.group(3), get_month(m.group(2)), int(m.group(1)))
                        if m
                        else v
                    )

            elif field.endswith(':hour'):
                regex = re.compile(r'(\d+):00 (\d+) (\w+)')

                def mapper(v, range):
                    if not v:
                        return v
                    date = range.get('from')
                    return date if date else v

            if mapper:
                raw_field = field.split(':', 1)[0]
                has_range = self.odoo.major_version >= 15
                for d in data:
                    if has_range:
                        d_range = d['__range'].get(raw_field)
                    else:
                        # parse the domain to get the range
                        d_range = {}
                        for e in d['__domain']:
                            if isinstance(e, list) and len(e) == 3 and e[0] == raw_field:
                                if e[1] == ">=" and 'from' not in d_range:
                                    d_range['from'] = e[2]
                                elif e[1] == "<" and 'to' not in d_range:
                                    d_range['to'] = e[2]
                    d[field] = mapper(d[field], d_range)
        return data

    def __read_dict_recursive(self, data, fields):
        """For each field, read recursively the data"""
        if not fields:
            fields = {f: {} for f in self.fields()}
        for field_name, child_fields in fields.items():
            field_info = self.fields().get(field_name, {})
            model_name = field_info.get('relation')
            if not model_name:
                # not a reference field, skip it
                continue

            # simplify contents and get ids
            many = field_info.get('type') != 'many2one'
            ids = set()
            if many:
                for datum in data:
                    value = datum.get(field_name)
                    if isinstance(value, list):
                        ids.update(value)
                    else:
                        datum[field_name] = []
            else:
                for datum in data:
                    value = datum.get(field_name)
                    if isinstance(value, list):
                        assert len(value) == 2 and not isinstance(value[1], int)
                        datum[field_name] = value[0]
                        ids.add(value[0])
            if not ids or not (set(child_fields) - {'id'}):
                continue

            # read the data from children
            model = self.odoo.get_model(model_name)
            children_data = model.read(list(ids), list(child_fields))
            model.__read_dict_recursive(children_data, child_fields)
            children_data = {e['id']: e for e in children_data}

            # replace the data
            if many:
                for datum in data:
                    datum[field_name] = [
                        children_data.get(v) or {"id": v} for v in datum.get(field_name) or []
                    ]
            else:
                for datum in data:
                    v = datum.get(field_name)
                    datum[field_name] = (children_data.get(v) or {"id": v}) if v else {}

        return data

    def read_dict(
        self,
        ids: Union[List[int], int],
        fields: Union[List[str], Dict[str, Dict]],
    ):
        """Read with a dictionnary output and hierarchy view

        Example: model.search_read_dict([], ['partner_id.name', 'name'])

        :param domain: The domain for the search
        :param fields: A list of fields (may contain chains f1.f2)
                       or a dict containing fields to read {field: {child_fields...}}
        :param kwargs: Other arguments passed to search_read (limit, offet, orderby, etc.)
        :return: A list of found objects
        """
        single = False
        if isinstance(ids, int):
            ids = [ids]
            single = True
        fields = self.__prepare_dict_fields(fields)
        data = self.read(ids, list(fields))
        result = self.__read_dict_recursive(data, fields)
        return result[0] if single else result

    def search_read_dict(self, domain: List, fields: Union[List[str], Dict[str, Dict]], **kwargs):
        """Search read with a dictionnary output and hierarchy view

        Similar to `read_dict`.

        :param domain: The domain for the search
        :param fields: A list of fields (may contain chains f1.f2)
                       or a dict containing fields to read {field: {child_fields...}}
        :param kwargs: Other arguments passed to search_read (limit, offet, orderby, etc.)
        :return: A list of found objects
        """
        fields = self.__prepare_dict_fields(fields)
        data = self.search_read(domain, list(fields), **kwargs)
        return self.__read_dict_recursive(data, fields)

    def read_group_dict(self, domain, aggregates, groupby, **kwargs):
        """Search read groupped data

        :param domain: The domain for the search
        :param aggregates: The aggregates
        :param groupby: Fields to group by
        :return: A list of groupped date
        """
        groupby = self.__prepare_dict_fields(groupby)
        groupby_list = list(groupby)
        if not groupby_list:
            raise ValueError('Missing groupby values')
        kwargs['lazy'] = False
        data = self.read_group(domain, aggregates or ['id'], groupby_list, **kwargs)
        data = self.__read_dict_date(data, groupby_list)
        return self.__read_dict_recursive(data, groupby)



import json
import logging
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Union, cast, overload

from .format import Formatter, decode_binary
from .odoo_rpc import OdooClient, OdooModel, urljoin

__doc__ = """Export and import data from Odoo.

## Data
To download values, use `export_data` which will return a table-like structure.
To import, you can use `load_data` to upload data into Odoo.
Alternatively, you can use the formatter to prepare data for upload and
then call create or write methods.

## Attachments
Get attachments and read binary values.

## Reports
Generate a report.
"""


#######################################
# ATTACHMENTS


@overload
def get_attachment(model: OdooModel, id: int) -> bytes:
    """Get attachment by ID"""
    ...


@overload
def get_attachment(model: OdooModel, *, encoded_value: str) -> bytes:
    """Get attachment from base64 encoded value"""
    ...


@overload
def get_attachment(model: OdooModel, id: int, field_name: str) -> bytes:
    """Get attachment from a field in an object by id"""
    ...


def get_attachment(
    model: OdooModel, id: int = 0, field_name: str = None, encoded_value: str = None
) -> bytes:
    if encoded_value is not None:
        return decode_binary(encoded_value)
    if not id:
        return b''
    # we read an attachment by id
    if not field_name:
        return get_attachments(model.odoo, [id])[id]
    # if we have a field name, decode the value or get the ID
    field_info = model.fields().get(field_name, {})
    record = model.read(id, [field_name])
    value = record[0][field_name]
    if field_info.get("relation") == "ir.attachment":
        if field_info.get("type") != "many2one":
            raise RuntimeError(
                "Field %s is not a many2one, here are the values: %s" % (field_name, value)
            )
        if not value:
            return b''
        value = value[0]
    elif field_info.get("type") != "binary":
        raise RuntimeError("%s is neither a binary or an ir.attachment field" % field_name)
    if isinstance(value, int):
        return get_attachment(model, value)
    return decode_binary(value)


def get_attachments(odoo: OdooClient, ids: List[int]) -> Dict[int, bytes]:
    """Get a list of tuples (name, raw_bytes) from ir.attachment by ids

    :param odoo: Odoo client
    :param ids: List of ids of attachments
    :return: Dict id -> raw_bytes
    """
    data = odoo['ir.attachment'].read(ids, ['datas'])
    return {r['id']: decode_binary(r['datas']) for r in data}


def list_attachments(
    model: OdooModel, ids: List[int], *, domain=[], fields=[], generate_access_token=False
) -> List[Dict]:
    """List attachments

    We search all attachments linked to the model and ids.
    All attachments, no matter if they have a res_field or not are read.
    We read the id, name, res_id, res_field.
    Special field names: public_url (generate public URL), datas (contents).

    :param model: The model
    :param ids: The record ids to filter on
    :param domain: Additional domain
    :param fields: Additional fields to read
    :param generate_access_token: Generate access token for the found documents
    :return: List of read properties from the attachments (id, name, res_field, ...)
    """
    url_field = None
    if 'public_url' in fields:
        url_field = 'public_url'
        fields += ['website_url', 'url', 'public', 'access_token']
    fields = list({'id', 'name', 'res_id', 'res_field', *fields} - {url_field})
    # Get the data
    if model.model == 'ir.attachment':
        attachments = model
        data = attachments.search_read(
            [
                ('id', 'in', ids),
            ]
            + domain,
            fields,
        )
    else:
        attachments = model.odoo['ir.attachment']
        data = attachments.search_read(
            [
                ('res_model', '=', model.model),
                ('res_id', 'in', ids),
                ('id', '!=', 0),  # to get all res_field
            ]
            + domain,
            fields,
        )
    # Get contents
    if 'datas' in fields:
        for d in data:
            d['datas'] = decode_binary(d['datas']) if d['datas'] else False
    # Generate access tokens
    if generate_access_token and data:
        tokens = attachments.execute('generate_access_token', [d['id'] for d in data])
        for d, token in zip(data, tokens):
            d['access_token'] = token
    # Compute a full url (false when not accessible)
    if url_field:

        def attachment_url(d, _model_name, _id):
            url = d.get('url') or d.get('website_url')
            if url and d.get('public'):
                return url
            if url and d.get('access_token'):
                url += ('&' if '?' in url else '?') + 'access_token=' + d.get('access_token')
                return url
            return False

        add_url(attachments, data, url_field=url_field, build_url=attachment_url)
    return data


def _download_content(
    odoo: OdooClient, url: str, *, params: Dict = {}, access_token: Optional[str] = None
) -> bytes:
    """Download contents from a URL"""
    # In the implementation, the session is authenticated.
    session = getattr(odoo, 'session', None)
    if not session:
        raise RuntimeError('Odoo client must have a session to download contents')

    url = urljoin(odoo.url, url)
    params = {}
    if access_token:
        params = {**params, 'access_token': access_token}
    req = session.get(url, params=params)
    req.raise_for_status()
    return req.content


def download_content(
    model: OdooModel,
    id: int,
    field_name: Optional[str] = None,
    *,
    access_token: Optional[str] = None,
) -> bytes:
    """Download contents from /web"""
    if field_name:
        url = f"/web/content/{model.model}/{id}/{field_name}"
    else:
        url = f"/web/content/{id}"
    return _download_content(model.odoo, url, access_token=access_token)


def download_image(
    model: OdooModel,
    id: int,
    field_name: Optional[str] = None,
    *,
    dimensions=None,
    access_token: Optional[str] = None,
):
    """Download an image from /web"""
    if field_name:
        url = f"/web/image/{model.model}/{id}/{field_name}"
    else:
        url = f"/web/image/{id}"
    if dimensions:
        url += "/%dx%d" % dimensions
    return _download_content(model.odoo, url, access_token=access_token)


#######################################
# REPORTS


def list_reports(model: OdooModel) -> List[Dict]:
    """List of reports from a model"""
    return model.odoo['ir.actions.report'].search_read(
        [('model', '=', model.model)], ['name', 'report_type', 'report_name']
    )


def get_report(model: OdooModel, report_name: str, id: int, converter='pdf') -> bytes:
    """Generate and download a report

    :param model: Odoo model
    :param reportname: The name of the report
    :param id: The id of the object
    :param converter: The output type (default: pdf)
    :return: Bytes of the report
    """
    # We need to use the session, because there is no public API since v14 to
    # render a report and get the bytes.
    if converter.startswith('qweb-'):
        converter = converter[5:]
    url = f"/report/{converter}/{report_name}/{id}"
    return _download_content(model.odoo, url)


#######################################
# IMPORT AND EXPORT


def make_batches(
    data: Iterable[Dict], *, batch_size: int = 1000, group_by: str = None
) -> Iterable[List[Dict]]:
    """Split an interable to a batched iterable

    :param data: The iterable
    :param batch_size: Target batch size (default: 1000)
    :param group_by: field to group by (the value is kept in a single batch
    :return: An iterable of batches
    """
    batch: List[Dict] = []
    if group_by:
        data = sorted(data, key=lambda d: d[group_by])
        group_value = None
        for d in data:
            current_value = d[group_by]
            if len(batch) >= batch_size and group_value != current_value:
                yield batch
                batch = [d]
            else:
                batch.append(d)
            group_value = current_value
    else:
        for d in data:
            batch.append(d)
            if len(batch) >= batch_size:
                yield batch
                batch = []
    if batch:
        yield batch


def load_data(
    model: OdooModel,
    data: Iterable,
    *,
    method: str = "load",
    method_row_type=None,
    fields: Optional[List[str]] = None,
    formatter: Formatter = None,
):
    """Load the data into the model.

    The method is executed given as parameter the formatted data
    and when row type is list, fields is also passed by keyword.
    If the row type is list and no fields are given, the first row is used.

    The default method used is `load` where field name's "." are replaced with
    "/" to make it work.
    `write` is handled specially, by making write() and create() calls.
    You can use `create` to insert new records.
    Otherwise, the model can have a custom method you could use.

    :param model: The model
    :param data: The list of rows (dict or list)
    :param method: The name of the method to call
    :param method_row_type: The type of the rows; dict (default) or list
    :param fields: List of field name for list rows
    :param formatter: The formatter to use
    :return: The result of the method call
    """
    log = logging.getLogger(__name__)

    if method == 'load':
        method_row_type = list
    elif method_row_type is None or method in ('create', 'write'):
        method_row_type = dict
    log.debug("Load: %s(), row type is %s", method, method_row_type)

    if formatter is None:
        formatter = Formatter(model)

    log.info("Load: convert and format data")
    if method_row_type == dict:
        data = __convert_to_type_dict(data, fields)
        data = [formatter.format_dict(d) for d in data]
    elif method_row_type == list:
        data, fields = __convert_to_type_list(data, fields)
        data = [[formatter.format_function[fields[i]](v) for i, v in enumerate(d)] for d in data]
    else:
        raise Exception('Unsupported method row type: %s' % method_row_type)

    log.info("Load data using %s.%s(), %d records", model.model, method, len(data))
    if method == 'load':
        fields = [f.replace('.', '/') for f in cast(List[str], fields)]
        return model.execute(method, fields=fields, data=data)
    if method == 'write':
        return __load_data_write(model, data)
    if fields:
        return model.execute(method, data, fields=fields)
    return model.execute(method, data)


def __convert_to_type_list(
    data: Iterable, fields: Optional[List[str]]
) -> Tuple[Iterable[List], List[str]]:
    idata = iter(data)
    first_row = next(idata, None)
    if first_row is None:
        return [], fields or ['id']
    if not isinstance(first_row, list):
        if not fields:
            raise RuntimeError('Missing fields to convert into type list')
        data = ([d.get(f) for f in fields] for d in data)
    elif fields is None:
        logging.getLogger(__name__).debug("Load: using first row as field names")
        fields = first_row
        data = idata
    elif not isinstance(data, list):
        data = [first_row] + list(data)
    return data, fields


def __convert_to_type_dict(data: Iterable, fields: Optional[List[str]]) -> Iterable[Dict]:
    idata = iter(data)
    first_row = next(idata, None)
    if first_row is None:
        return []
    if isinstance(first_row, list):
        if fields is None:
            fields = first_row
            data = (dict(zip(fields, d)) for d in idata)
        else:
            if not isinstance(data, list):
                data = [first_row] + list(idata)
            if fields:
                data = ({f: d[i] for i, f in enumerate(fields)} for d in data)
    elif not isinstance(data, list):
        data = [first_row] + list(idata)
    return data


def __load_data_write(model: OdooModel, data: List[Dict]) -> Dict:
    """Use multiple write() and create() calls to update data

    :return: {'write_count': x, 'create_count': x, 'ids': [list of ids]}
    """
    create_data = []
    ids = []
    for d in data:
        id = d.pop('id')
        if id:
            model.execute('write', id, d)
            ids.append(id)
        else:
            create_data.append(d)
            ids.append(0)
    if create_data:
        created_ids = model.execute('create', create_data)
        iids = iter(created_ids)
        for i in range(len(ids)):
            if not ids[i]:
                ids[i] = next(iids)
        assert next(iids, None) is None
    return {
        'write_count': len(data) - len(create_data),
        'create_count': len(create_data),
        'ids': ids,
    }


def export_data(
    model: OdooModel,
    filter_or_domain: Union[str, List],
    export_or_fields: Union[str, List[str]],
    with_header: bool = True,
    expand_many: bool = False,
) -> List[List]:
    """Export data into a tabular format

    :param model: Odoo model
    :param filter_or_domain: Either a domain or an ir.filer name
    :param export_or_fields: Either a list of fields (like is search_read_dict)
        or an ir.exports name
    :param with_header: Include the header in the result (default: True)
    :param expand_many: Flatten lists embedded in the result (default: False)
    :return: List of rows with data
    """
    log = logging.getLogger(__name__)
    odoo = model.odoo
    kwargs = {}

    if isinstance(filter_or_domain, str):
        log.debug('Get the filter from Odoo: %s', filter_or_domain)
        filter_data = odoo['ir.filters'].search_read_dict(
            [
                ('name', '=', filter_or_domain),
                ('model_id', '=', model.model),
            ],
            ['domain', 'context', 'sort'],
        )
        if not filter_data:
            raise ValueError('Filter not found')
        domain = filter_data.get('domain')
        if filter_data.get('sort'):
            kwargs['order'] = json.loads(filter_data.get('sort'))
        if filter_data.get('context'):
            kwargs['context'] = json.loads(filter_data.get('context'))
    else:
        domain = filter_or_domain

    if isinstance(export_or_fields, str):
        log.debug('Get the export lines: %s', export_or_fields)
        fields = odoo['ir.exports.line'].search_read(
            [
                ('export_id.name', '=', export_or_fields),
                ('export_id.resource', '=', model.model),
            ],
            ['name'],
        )
        fields = [f['name'].replace('/', '.') for f in fields]
    else:
        log.debug('List of fields to export: %s', export_or_fields)
        fields = export_or_fields
    if not fields:
        raise ValueError('No fields to export')

    log.info('Export: execute search on %s', model.model)
    records = model.search_read_dict(domain, fields)

    log.info('Export: done, cleaning data')
    data = flatten(records, fields, expand_many=expand_many)
    if with_header:
        data.insert(0, [str(f) for f in fields])
    return data


def add_fields(
    model: OdooModel, data: List[Dict], by_field: str, fields: List[str] = ['id'], domain: List = []
):
    """Add fields by querying the model

    :param model: The model
    :param data: The list of dicts
    :param by_field: The field name to use to find other fields
    :param fields: The fields to get (default: ['id'])
    :param domain: Additional domain to use
    :raises Exception: When multiple results have the same by_field key
    """
    domain_by_field: List[Any] = [(by_field, 'in', [d[by_field] for d in data])]
    domain = ['&'] + domain_by_field + (domain if domain else domain_by_field)
    fetched_data = model.search_read_dict(domain, fields + [by_field])
    index = {d.pop(by_field): d for d in fetched_data}
    if len(index) != len(fetched_data):
        raise Exception('%s is not unique in %s when adding fields' % (by_field, model.model))
    for d in data:
        updates = index.get(d[by_field])
        if updates:
            d.update(updates)
        else:
            d.update({f: None for f in fields})


def add_xml_id(model: OdooModel, data: List, *, id_name='id', xml_id_field='xml_id'):
    """Add a field containg the xml_id

    :param model: The model
    :param data: The list of dicts or rows (list to append to)
    :param id_name: The name of the field to get the resource ids (default: 'id')
    :param xml_id_field: The name of the xmlid field (default: 'xml_id')
    """
    if id_name != 'id':
        relation = model.fields()[id_name].get('relation')
        model = model.odoo.get_model(cast(str, relation), check=True)
    id_index: Optional[int] = None
    ids = set()
    for row_num, row in enumerate(data):
        if isinstance(row, dict):
            ids.add(row[id_name])
        elif isinstance(row, list):
            if row_num == 0:
                id_index = row.index(id_name)
            else:
                ids.add(row[cast(int, id_index)])
        else:
            raise TypeError('Cannot append the url to %s' % type(row))
    xml_ids = {
        i['res_id']: i['complete_name']
        for i in model.odoo['ir.model.data'].search_read(
            [('model', '=', model.model), ('res_id', 'in', list(ids))],
            ['complete_name', 'res_id'],
        )
    }
    if id_index is None:
        # we have dicts
        for row in data:
            row[xml_id_field] = xml_ids.get(row[id_name], False)
    else:
        # we have a list
        for row_num, row in enumerate(data):
            if row_num == 0:
                row.append(xml_id_field)
            else:
                row.append(xml_ids.get(row[id_index], False))


def add_url(
    model: OdooModel,
    data: List,
    *,
    url_field='url',
    model_id_func=None,
    build_url=lambda _r, model_name, id: f"/web#model={model_name}&id={id}",
):
    """Add an URL field to data

    :param model: The model
    :param data: The list of dicts or rows (list to append to)
    :param url_field: The name of the URL field (default: 'url')
    :param model_id_func: The function to get a tuple (model_name, id)
    :param build_url: Function to build the URL (record, model_name, id) -> "/web..."
    """
    base_url = model.odoo.url
    if not model_id_func and data:
        if isinstance(data[0], list):
            model_id_func = lambda r: (model.model, r[0])  # noqa: E731
        else:
            model_id_func = lambda d: (model.model, d['id'])  # noqa: E731
    for row_num, row in enumerate(data):
        model_name, id = model_id_func(row)
        url = build_url(row, model_name, id)
        if url:
            url = urljoin(base_url, url)
        if isinstance(row, dict):
            row[url_field] = url
        elif isinstance(row, list):
            if row_num == 0:
                row.append(url_field)
            else:
                row.append(url)
        else:
            raise TypeError('Cannot append the url to %s' % type(row))


@overload
def _flatten(value: List, field_levels: List[str]) -> List:
    ...


@overload
def _flatten(value: Dict, field_levels: List[str]) -> Union[Any, bool]:
    ...


@overload
def _flatten(value: Any, field_levels: List[str]) -> Any:
    ...


def _flatten(value, field_levels: List[str]) -> Any:
    if not field_levels:
        return value
    if isinstance(value, list):
        return [_flatten(i, field_levels) for i in value]
    if isinstance(value, dict):
        return _flatten(value.get(field_levels[0]), field_levels[1:])
    return False  # default


def _expand_many(data: List[Dict]) -> Iterator[Dict]:
    for d in data:
        should_yield = True
        for k, v in d.items():
            if isinstance(v, list) and len(v):
                should_yield = False
                for i in v:
                    yield from _expand_many([{**d, k: i}])
        for k, v in d.items():
            if should_yield and isinstance(v, dict):
                should_yield = False
                for row in _expand_many([v]):
                    yield {**d, k: row}
        if should_yield:
            yield d


def flatten(data: List[Dict], fields: List[str], expand_many=False) -> List[List]:
    """Flatten each dict with values into a single row"""
    if expand_many:
        data = list(_expand_many(data))
    field_levels = [f.split('.') for f in fields]
    return [[_flatten(d.get(fl[0]), fl[1:]) for fl in field_levels] for d in data]


__all__ = [
    'get_attachment',
    'get_attachments',
    'list_attachments',
    'get_report',
    'list_reports',
    'Formatter',
    'make_batches',
    'load_data',
    'export_data',
    'add_fields',
    'add_url',
    'add_xml_id',
]



import logging
import re

import pytest
import requests

import odoo_connect


@pytest.fixture(scope='session')
def connect_params():
    runbot_page = requests.get('https://runbot.odoo.com/')
    odoo_url = "http://localhost/"
    version = ""
    # we don't use (yet) any XML parsing to keep dependencies simple
    # find one of the following and select a URL
    # - bundle href (row and name of the branch)
    # - link to a sign-in action
    for m in re.finditer(
        r'href="/runbot/(bundle)[^<]+<b>([^<]*)</b>'
        r'|<a class="[^"]*fa-(sign-in)[^"]*" href="([^"]*)"',
        runbot_page.text,
    ):
        if m.group(1) == 'bundle':
            version = m.group(2)
        elif version and re.match(r'^\d+\.\d$', version) and m.group(3) == 'sign-in':
            odoo_url = m.group(4)
            break
    odoo_url = odoo_url.replace('http://', 'https://')
    logging.info("Using odoo server %s: %s", version, odoo_url)

    return {
        'url': odoo_url,
        'username': 'admin',
        'password': 'admin',
    }


@pytest.fixture(scope='session')
def odoo_session(connect_params):
    return odoo_connect.connect(**connect_params)



import pytest

import odoo_connect.data as odoo_data


def test_add_field(odoo_cli, odoo_json_rpc_handler):
    handler = odoo_json_rpc_handler

    @handler.patch_execute_kw('res.partner', 'read')
    def read_partner(id, fields=[]):
        return [{'id': 1, 'name': 'test'}]

    @handler.patch_execute_kw('res.partner', 'search_read')
    def read_search_partner(domain, fields=[]):
        print(domain)
        data = [{'id': 1, 'name': 'test'}]
        if not fields:
            return data
        if 'id' not in fields:
            fields += 'id'
        return [{k: v for k, v in d.items() if k in fields} for d in data]

    @handler.patch_execute_kw('res.partner', 'fields_get')
    def read_fields_partner(allfields=[], attributes=[]):
        attr = {a: True if a == 'store' else False for a in attributes}
        return {
            'id': {**attr, 'type': 'int', 'string': 'ID'},
            'name': {**attr, 'type': 'char', 'string': 'Name'},
        }

    model = odoo_cli['res.partner']
    data = model.read(1, ['name'])
    id = data[0].pop('id')
    odoo_data.add_fields(model, data, 'name', ['id'])
    assert data[0].get('id') == id


def test_add_url(odoo_cli):
    model = odoo_cli['res.partner']
    data = [{'id': 4}]
    odoo_data.add_url(model, data)
    url = data[0]['url']
    assert url.startswith('http')
    assert 'model=res.partner' in url
    assert 'id=4' in url


def test_add_url_list(odoo_cli):
    model = odoo_cli['res.partner']
    data = [['id'], [4]]
    odoo_data.add_url(model, data)
    assert 'url' in data[0]
    assert model.model in data[1][1]


def test_add_xml_id(odoo_cli, odoo_json_rpc_handler):
    handler = odoo_json_rpc_handler

    @handler.patch_execute_kw('ir.model.data', 'search_read')
    def read_search_data(domain, fields=[]):
        print(domain)
        data = [{'id': 1, 'res_id': 4, 'model': 'res.partner', 'complete_name': 'test.myid'}]
        if not fields:
            return data
        if 'id' not in fields:
            fields += 'id'
        return [{k: v for k, v in d.items() if k in fields} for d in data]

    model = odoo_cli['res.partner']
    data = [{'id': 4}, {'id': 9}]
    odoo_data.add_xml_id(model, data)
    assert 'test.myid' == data[0]['xml_id']
    assert not data[1]['xml_id']


@pytest.mark.parametrize(
    "dict,fields,expected",
    [
        ({'id': 5}, ['id'], [5]),
        ({'id': 5, 'name': 'test'}, ['name'], ['test']),
        (
            {'id': 5, 'partner_id': {'name': 'partner_name'}},
            ['id', 'partner_id.name'],
            [5, 'partner_name'],
        ),
    ],
)
def test_flatten(dict, fields, expected):
    output = odoo_data.flatten([dict], fields)[0]
    assert expected == output


@pytest.mark.parametrize(
    "list_dict,fields,expected",
    [
        ([], ['id'], []),
        (
            [{'partner_id': {'name': 'nn'}, 'order_id': [{'name': 'S1'}, {'name': 'S2'}]}],
            ['partner_id.name', 'order_id.name'],
            [['nn', ['S1', 'S2']]],
        ),
    ],
)
def test_flatten_many(list_dict, fields, expected):
    output = odoo_data.flatten(list_dict, fields)
    assert expected == output


def test_flatten_expand_many():
    expanded = list(
        odoo_data.flatten(
            [
                {
                    'id': 5,
                    'numbers': [2, False],
                    'orders': [{'name': 'ok'}, {'name': 'ko', 'partners': [1, 2]}],
                }
            ],
            ['numbers', 'orders.name', 'orders.partners'],
            expand_many=True,
        )
    )
    assert len(expanded) == 12



from pprint import pprint

import odoo_connect.data as odoo_data


def test_export(odoo_session):
    model = odoo_session['res.partner']
    fields = ['name', 'parent_id.name']
    data = odoo_data.export_data(model, [('id', '<', 10)], fields)
    pprint(data)
    assert fields == data[0]
    assert all(len(fields) == len(row) for row in data)



import re


def test_read_dict(odoo_session):
    data = odoo_session['res.users'].read_dict(
        [1],
        ['login', 'partner_id.name', 'partner_id.commercial_partner_id.name'],
    )
    print(data)
    user = data[0]
    assert user['id'] == 1
    assert 'login' in user
    partner = user['partner_id']
    assert 'id' in partner
    assert isinstance(partner.get('commercial_partner_id'), dict)


def test_read_one_dict(odoo_session):
    user = odoo_session['res.users'].read_dict(
        1,
        ['login'],
    )
    assert isinstance(user, dict)
    assert 'login' in user


def test_search_read_dict(odoo_session):
    data = odoo_session['res.users'].search_read_dict(
        [('id', '=', 1), ('active', 'in', [True, False])],
        ['login', 'partner_id.name', 'partner_id.commercial_partner_id.name'],
    )
    print(data)
    user = data[0]
    assert user['id'] == 1
    assert 'login' in user
    partner = user['partner_id']
    assert 'id' in partner
    assert isinstance(partner.get('commercial_partner_id'), dict)


def test_read_all_fields(odoo_session):
    user = odoo_session['res.users'].read_dict(1, [])
    assert isinstance(user.get('partner_id'), int)


def test_read_group(odoo_session):
    data = odoo_session['res.users'].read_group_dict([], [], ['create_date:month'])
    assert data[0]['__count']
    month = data[0]['create_date:month']
    assert re.match(r'\d{4}-\d{2}', month), 'Date is not in ISO format'



import odoo_connect.data as odoo_data


def test_list_reports(odoo_session):
    model = odoo_session['account.move']
    reports = odoo_data.list_reports(model)
    assert all(d.get('report_name') for d in reports)


def test_get_report(odoo_session):
    model = odoo_session['account.move']
    value = odoo_data.get_report(model, 'account.report_invoice', 1)
    assert isinstance(value, bytes) and len(value)



import odoo_connect.data as odoo_data
from odoo_connect.format import format_binary


def test_get_attachment(odoo_session):
    model = odoo_session['account.move']
    value = odoo_data.get_attachment(model, 1)
    assert isinstance(value, bytes) and len(value) > 1


def test_get_attachment_field(odoo_session):
    model = odoo_session['res.users']
    ids = model.search([('image_1920', '!=', False)], limit=1)
    value = odoo_data.get_attachment(model, ids[0], 'image_1920')
    assert isinstance(value, bytes) and len(value) > 1


def test_get_attachment_bytes(odoo_session):
    bin = bytes("hello", 'ascii')
    value = odoo_data.get_attachment(odoo_session['res.partner'], encoded_value=format_binary(bin))
    assert bin == value


def test_get_attachments(odoo_session):
    values = odoo_data.get_attachments(odoo_session, [1, 2])
    assert isinstance(values, dict) and values



import math
from datetime import date, datetime

import pytest

import odoo_connect.format as odoo_format


def test_binary_encoding():
    bin = b'OK'
    value = odoo_format.format_binary(bin)
    assert isinstance(value, str)
    assert len(value) == math.ceil(len(bin) / 3) * 4

    original = odoo_format.decode_binary(value)
    assert isinstance(original, bytes) and original == bin


def test_binary_encoding_empty():
    assert odoo_format.format_binary(b'') == ''


@pytest.mark.parametrize(
    "type_name,func,input,expected",
    [
        ('char', 'default', 'test', 'test'),
        ('int', 'default', 3, 3),
        ('date', 'date', datetime(2020, 2, 2, 3), "2020-02-02"),
        ('date', 'date', date(2020, 2, 2), "2020-02-02"),
        ('date', 'date', None, False),
        ('datetime', 'datetime', datetime(2020, 2, 2, 3, microsecond=3), "2020-02-02 03:00:00"),
        ('datetime', 'datetime', "2020-02-02 03:00:00.3", "2020-02-02 03:00:00"),
        ('datetime', 'datetime', date(2020, 2, 2), "2020-02-02 00:00:00"),
        ('datetime', 'datetime', None, False),
        ('binary', 'binary', b'', ''),
        ('char', 'default', '', False),
    ],
)
def test_format(type_name, func, input, expected):
    formatter = getattr(odoo_format, "format_%s" % func)
    assert formatter(input) == expected, "Couldn't format %s" % type_name


@pytest.mark.parametrize(
    "type_name,func,input,expected",
    [
        ('char', 'default', 'test', 'test'),
        ('int', 'default', 3, 3),
        ('date', 'date', "2022-02-02", date(2022, 2, 2)),
        ('date', 'date', False, None),
        ('datetime', 'datetime', "2020-02-02 03:00:00", datetime(2020, 2, 2, 3)),
        ('binary', 'binary', '', b''),
    ],
)
def test_decode(type_name, func, input, expected):
    decoder = getattr(odoo_format, "decode_%s" % func)
    assert decoder(input) == expected, "Couldn't decode %s" % type_name


def test_formatter():
    f = odoo_format.Formatter()
    f.format_function['d'] = odoo_format.format_date
    f.field_map['X'] = 'y'
    f.field_map['removed'] = ''
    assert f.map_field_name('X') == 'y'
    d = f.format_dict({'d': '2022-01-01 15:10:05', 'X': 'value y', 'def': 'ok ', 'removed': 1})
    print(d)
    assert d['d'] == '2022-01-01'
    assert 'X' not in d and d['y'] == 'value y'
    assert d['def'] == 'ok'
    assert 'removed' not in d


def test_decoder():
    f = odoo_format.Formatter()
    f.decode_function['d'] = odoo_format.decode_date
    d = f.decode_dict({'d': '2022-01-01', 'x': '2022-01-01'})
    print(d)
    assert d['d'] == date(2022, 1, 1)
    assert d['x'] == '2022-01-01'



import itertools

import odoo_connect.data as odoo_data


def test_make_batches():
    numbers = range(10)
    bnumbers = list(odoo_data.make_batches(numbers, batch_size=3))
    assert len(bnumbers) == 4


def test_make_batches_by_key():
    data = [
        {'letter': letter, 'i': i} for letter, i in zip(itertools.cycle(['a', 'b', 'c']), range(10))
    ]
    batches = odoo_data.make_batches(data, batch_size=5, group_by='letter')
    batch = next(batches)
    assert len(batch) == 7  # 4 a and 3 b
    batch = next(batches)
    assert len(batch) == 3  # 3 c
    assert next(batches, None) is None


def test_import(odoo_session):
    model = odoo_session['res.partner']
    data = [{'name': 'john'}]
    ids = odoo_data.load_data(model, data, method='create')
    assert len(ids) == 1


def test_import_create_rows(odoo_session):
    model = odoo_session['res.partner']
    data = [['name'], ['john row']]
    ids = odoo_data.load_data(model, data, method='create')
    assert len(ids) == 1



from typing import Any, Callable, Dict, Iterable, List, Union

import odoo_connect.format

from . import odoo_rpc

__doc__ = """Interact more easily with Odoo records."""


"""Cache of read values"""
GLOBAL_CACHE: Dict[odoo_rpc.OdooModel, Dict[int, Dict[str, Any]]] = {}


class Instance:
    """A proxy for an instance set"""

    __model: odoo_rpc.OdooModel
    __ids: List[int]

    def __init__(self, model: odoo_rpc.OdooModel, ids: List[int]) -> None:
        self.__model = model
        self.__ids = ids

    def __bool__(self) -> bool:
        return bool(self.__ids)

    def __len__(self):
        return len(self.__ids)

    def __dir__(self) -> Iterable[str]:
        return self.__model.fields().keys()

    def __add__(self, other) -> "Instance":
        if self.__model.model != other.__model.model or not isinstance(other, Instance):
            raise ValueError('Cannot combine different models')
        ids = self.__ids + other.__ids
        return Instance(self.__model, ids)

    def __sub__(self, other) -> "Instance":
        if self.__model.model != other.__model.model or not isinstance(other, Instance):
            raise ValueError('Cannot combine different models')
        otherset = set(other.__ids)
        ids = [i for i in self.__ids if i not in otherset]
        return Instance(self.__model, ids)

    def __or__(self, other) -> "Instance":
        if self.__model.model != other.__model.model or not isinstance(other, Instance):
            raise ValueError('Cannot combine different models')
        ids = self.__ids + (other - self).__ids
        return Instance(self.__model, ids)

    def __and__(self, other) -> "Instance":
        if self.__model.model != other.__model.model or not isinstance(other, Instance):
            raise ValueError('Cannot combine different models')
        otherset = set(other.__ids)
        ids = [i for i in self.__ids if i in otherset]
        return Instance(self.__model, ids)

    @property
    def ids(self) -> List[int]:
        return self.__ids

    @property
    def _model(self) -> odoo_rpc.OdooModel:
        return self.__model

    def _formatter(self) -> odoo_connect.format.Formatter:
        return odoo_connect.format.get_default_formatter(self._model)

    def __getitem__(self, item) -> "Instance":
        ids = self.__ids[item]
        if not isinstance(ids, Iterable):
            ids = [ids]
        return Instance(self.__model, ids)

    def __getattr__(self, __name: str) -> Any:
        value = self._mapped(__name)
        if isinstance(value, list):
            if len(self.__ids) == 1:
                # decode the value get getting the attribute
                return self._formatter().decode_function[__name](value[0])
            if len(self.__ids) == 0:
                return False
            raise ValueError('Too many values to unpack: ' + __name)
        return value

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name.startswith('_Instance__'):
            return super().__setattr__(__name, __value)
        format = True
        if isinstance(__value, Instance):
            format = False
            prop = self.__model.fields().get(__name) or {}
            if '2many' in (prop.get('type') or ''):
                if len(__value) == 0:
                    __value = [(5, 0, 0)]  # remove all
                else:
                    __value = [(6, 0, __value.ids)]  # replace all
            else:
                if len(__value) > 1:
                    raise ValueError('Cannot assign multiple values')
                if len(__value) == 1:
                    __value = __value.ids[0]
                else:
                    __value = False
        # format the value when writing a scalar
        return self.write({__name: __value}, format=format)

    def mapped(self, path: str):
        """Map/read a field path"""
        paths = path.split('.', 1)
        if len(paths) > 1:
            value = self._mapped(paths[0])
            if not isinstance(value, Instance):
                raise ValueError(f'{paths[0]} is not a relation')
            return value.mapped(paths[1])
        return self._mapped(path)

    def _mapped(self, field_name: str) -> Union["Instance", List[Any]]:
        """Map/read a field"""
        prop = self.__model.fields().get(field_name)
        if not prop:
            raise ValueError(f'Invalid field: {field_name}')
        values = [d[field_name] for d in self.read(check_fields=[field_name])]
        relation = prop.get('relation')
        if relation:
            model = self.__model.odoo.get_model(relation)
            # value is either list[int] (many) or list[tuple[int, str]|False] (one)
            ids = set(i for v in values for i in (v or []) if isinstance(i, int) and i)
            return Instance(model, list(ids))
        return values

    def cache(self, fields: List[str] = [], computed=False, exists=False) -> "Instance":
        """Cache the record fields and return self"""
        fieldset = set(self._default_fields(computed=computed) + fields)
        model_cache = self.__cache()
        # find missing ids, when missing in cache or field missing in cache
        # read all at once to have more consistency and avoid roundtrips
        missing_ids = set(i for i in self.__ids if fieldset - model_cache.get(i, {}).keys())
        # an exists() check is not needed because read() will return only existing rows
        if not missing_ids:
            return self
        for d in self.__model.read(list(missing_ids), list(fieldset)):
            id = d['id']
            if id in model_cache:
                model_cache[id].update(d)
            else:
                model_cache[id] = d
        return self

    def read(self, *, check_fields: List[str] = []) -> List[Dict[str, Any]]:
        """Read the data"""
        self.cache(fields=check_fields)
        model_cache = self.__cache()
        try:
            return [model_cache[i] for i in self.__ids]
        except KeyError as e:
            raise odoo_rpc.OdooServerError(f"Cannot read {self.__model.model}: {e}")

    def _default_fields(self, *, computed=False) -> List[str]:
        """List of fields to read by default"""
        data = self.__model.fields()
        return [
            f
            for f, prop in data.items()
            if '2many' not in (prop.get('type') or '')
            and prop.get('type') != 'binary'
            and (computed or f in ('id', 'display_name') or prop.get('store'))
        ]

    def fields_get(self, field_names=[]) -> Dict[str, Dict]:
        """Get the field information"""
        data = self.__model.fields()
        if field_names:
            data = {k: v for k, v in data.items() if k in field_names}
        return data

    def browse(self, *ids: int) -> "Instance":
        """Create an instance with the given ids"""
        return Instance(self.__model, list(ids))

    def exists(self) -> "Instance":
        """Return only existing records"""
        # re-read records to validate
        self.invalidate_cache(self.__ids)
        self.cache(computed=False, exists=True)
        model_cache = self.__cache()
        ids = set(self.__ids) & model_cache.keys()
        return self.browse(*ids) if len(ids) < len(self.__ids) else self

    def search(self, domain: List, **kw) -> "Instance":
        """Search for an instance"""
        fields = self._default_fields()
        data = self.__model.search_read(domain, fields, **kw)
        # add only new data, keep cache consistent
        model_cache = self.__cache()
        model_cache.update({d['id']: d for d in data if d['id'] not in model_cache})
        return Instance(self.__model, [d['id'] for d in data])

    def name_search(self, name: str, **kw) -> "Instance":
        """Search by name"""
        # search and return only the ids
        data = self.__model.name_search(name, **kw)
        return Instance(self.__model, [d[0] for d in data])

    def create(self, *values: Dict[str, Any], format: bool = False) -> "Instance":
        """Create multiple instances"""
        if not values:
            return self.browse()
        if format:
            formatter = self._formatter().format_dict
            value_list = [formatter(d) for d in values]
        else:
            value_list = list(values)
        ids = self.__model.create(value_list)
        return self.browse(*ids)

    def write(self, values: Dict[str, Any], format: bool = False):
        """Update the values of the current instance"""
        if not values:
            return
        if format:
            values = self._formatter().format_dict(values)
        self.__model.write(self.__ids, values)
        self.invalidate_cache(self.__ids)

    def unlink(self):
        """Remove the records from the database"""
        self.invalidate_cache(self.__ids)
        self.__model.unlink(self.__ids)

    def copy(self):
        """Copy the records in the database"""
        ids = self.__model.copy(self.__ids)
        return self.browse(*ids)

    def __cache(self) -> Dict[int, Dict[str, Any]]:
        model_cache = GLOBAL_CACHE.get(self.__model)
        if not model_cache:
            GLOBAL_CACHE[self.__model] = model_cache = {}
        return model_cache

    def invalidate_cache(self, ids=None):
        """Invalidate the cache for a set of ids or all the model"""
        model_cache = GLOBAL_CACHE.get(self.__model)
        if not model_cache:
            return
        if ids is None:
            model_cache.clear()
        else:
            for id in ids:
                model_cache.pop(id, None)

    def filtered(self, predicate: Callable[["Instance"], bool]) -> "Instance":
        """Filter the records"""

        def _predicate(i):
            return predicate(self.browse(i))

        self.cache(computed=False)
        ids = list(filter(_predicate, self.__ids))
        return Instance(self.__model, ids)

    def sorted(self, order: Callable[["Instance"], Any]) -> "Instance":
        """Sort the objects by a field"""

        def sorted_key(i):
            return order(self.browse(i))

        self.cache(computed=False)
        ids = sorted(self.__ids, key=sorted_key)
        return Instance(self.__model, ids)

    def get_attachments(self) -> "Instance":
        """Return ir.attachment linked to this instance"""
        return explore(self.__model.odoo['ir.attachment']).search(
            [
                ('res_model', '=', self.__model.model),
                ('res_id', 'in', self.__ids),
                ('id', '!=', 0),  # to get all res_field
            ]
        )

    def _call(self, method, *args, model_method=False, **kw):
        """Call a method on the model

        :param model_method: Whether to don't pass ids
        """
        if model_method:
            return self.__model.execute(method, *args, **kw)
        return self.__model.execute(method, self.ids, *args, **kw)

    def __repr__(self) -> str:
        return repr(self.__model) + str(self.__ids)


def explore(model: odoo_rpc.OdooModel) -> Instance:
    """Create an empty instance to explore"""
    return Instance(model, [])


__all__ = ['explore']



import pytest

from odoo_connect.explore import Instance, explore


@pytest.fixture(scope='function')
def odoo_cli_partner(odoo_cli, odoo_json_rpc_handler) -> Instance:
    handler = odoo_json_rpc_handler
    data = [{'id': 1, 'name': 'test'}, {'id': 2, 'name': 'demo'}]
    for d in data:
        d['display_name'] = d['name']

    def read_partner(ids, fields=[]):
        if not fields:
            fields = data[0].keys()
        elif 'id' not in fields:
            fields = ['id'] + fields
        return [{k: v for k, v in d.items() if k in fields} for d in data if d['id'] in ids]

    handler.patch_execute_kw('res.partner', 'read')(read_partner)

    @handler.patch_execute_kw('res.partner', 'search_read')
    def read_search_partner(domain, fields=[]):
        print('read_search_partner', domain)
        return read_partner([1], fields=fields)

    @handler.patch_execute_kw('res.partner', 'fields_get')
    def read_fields_partner(allfields=[], attributes=[]):
        attr = {a: False for a in attributes}
        attr['store'] = True
        return {
            'id': {**attr, 'type': 'int', 'string': 'ID'},
            'name': {**attr, 'type': 'char', 'string': 'Name'},
            'display_name': {**attr, 'type': 'char', 'string': 'Display Name', 'store': False},
        }

    @handler.patch_execute_kw('res.partner', 'create')
    def create(val_list):
        id = max(d['id'] for d in data)
        result = []
        for d in val_list:
            id += 1
            d['id'] = id
            data.append(d)
            result.append(id)
        return result

    @handler.patch_execute_kw('res.partner', 'write')
    def write(ids, values):
        for d in data:
            if d['id'] in ids:
                d.update(values)
        return True

    partner = explore(odoo_cli['res.partner'])
    partner.invalidate_cache()
    return partner


def test_explore(odoo_cli):
    obj = explore(odoo_cli['res.partner'])
    assert obj._model.model == 'res.partner'
    assert not obj
    obj = obj.browse(9)
    assert obj
    assert obj.ids == [9]


def test_ex_read(odoo_cli_partner: Instance):
    inst = odoo_cli_partner
    inst = inst.browse(1)
    data = inst.read()
    assert data[0]['name'] == 'test'
    assert inst.name == 'test'


def test_ex_search(odoo_cli_partner: Instance):
    inst = odoo_cli_partner.search([('name', '=', 'test')])
    assert len(inst) == 1


def test_ex_cache(odoo_cli_partner: Instance):
    inst = odoo_cli_partner.browse(1).cache(['display_name'])
    assert isinstance(inst, Instance)
    assert inst.display_name


def test_ex_exist(odoo_cli_partner: Instance):
    inst = odoo_cli_partner.browse(*range(10)).exists()
    assert 0 < len(inst) < 10


def test_ex_combine(odoo_cli_partner: Instance):
    a = odoo_cli_partner.browse(5, 6)
    b = odoo_cli_partner.browse(6, 7)
    assert len(a + b) == 4
    assert (a - b).ids == [5]
    assert (a | b).ids == [5, 6, 7]
    assert (a & b).ids == [6]


def test_ex_filtered(odoo_cli_partner: Instance):
    inst = odoo_cli_partner.browse(1, 2)
    assert len(inst) == 2
    inst = inst.filtered(lambda i: i.name == 'test')
    assert len(inst) == 1


def test_ex_create(odoo_cli_partner: Instance):
    inst = odoo_cli_partner.create({'name': 'ok'})
    assert inst
    assert inst.name == 'ok'


def test_ex_write(odoo_cli_partner: Instance):
    inst = odoo_cli_partner.browse(1)
    assert inst.name == 'test'
    inst.name = 'ok'
    assert inst.name == 'ok'


def test_ex_write_format(odoo_cli_partner: Instance):
    inst = odoo_cli_partner.browse(1)
    assert inst.name == 'test'
    inst.write({'name': ''}, format=True)
    assert inst.name is False
